from typing import Any, Callable, Mapping, Sequence, Type
from jinja2.defaults import (
    BLOCK_END_STRING,
    BLOCK_START_STRING,
    COMMENT_END_STRING,
    COMMENT_START_STRING,
    KEEP_TRAILING_NEWLINE,
    LSTRIP_BLOCKS,
    TRIM_BLOCKS,
    VARIABLE_END_STRING,
    VARIABLE_START_STRING,
)
from jinja2.ext import Extension
from starlette.background import BackgroundTask
from starlette.requests import Request
from pydantic import BaseModel, field_validator, Field
from pypox.processing.base import HeaderValidator
from jinja2 import BaseLoader, BytecodeCache, Environment, FileSystemLoader, Undefined
from starlette.responses import HTMLResponse


class HTMXValidator(HeaderValidator):

    def __init__(self, name: str, request: Request) -> None:
        super().__init__(name, request)

    def __call__(self, _type: Type) -> Any:
        if _type not in [HTMXHeaders]:
            return {}
        return {self._name: self.validate()}

    def validate(self) -> Any:
        data = {
            key: val
            for key, val in self._request.headers.items()
            if key.startswith("hx-")
        }
        return HTMXHeaders.model_validate(data)


class HTMXHeaders(BaseModel):

    boosted: str = Field(default="false", alias="hx-boosted")
    current_url: str = Field(default="", alias="hx-current-url")
    history_restored: str = Field(default="false", alias="hx-history-restored")
    prompt: str = Field(default="false", alias="hx-prompt")
    request: str = Field(default="false", alias="hx-request")
    target: str = Field(default="", alias="hx-target")
    trigger_name: str = Field(default="", alias="hx-trigger-name")
    trigger: str = Field(default="", alias="hx-trigger")

    @field_validator("request")
    def validate_request(cls, value):
        if value:
            return value
        else:
            raise ValueError("Request must be True")


class HTMXResponseHeaders(BaseModel):
    location: str = Field(
        default="", validation_alias="hx-location", serialization_alias="hx-location"
    )
    push_url: str = Field(
        default="", validation_alias="hx-push-url", serialization_alias="hx-push-url"
    )
    redirect: str = Field(
        default="", validation_alias="hx-redirect", serialization_alias="hx-redirect"
    )
    refresh: str = Field(
        default="", validation_alias="hx-refresh", serialization_alias="hx-refresh"
    )
    replace_url: str = Field(
        default="",
        validation_alias="hx-replace-url",
        serialization_alias="hx-replace-url",
    )
    reswap: str = Field(
        default="", validation_alias="hx-reswap", serialization_alias="hx-reswap"
    )
    retarget: str = Field(
        default="", validation_alias="hx-retarget", serialization_alias="hx-retarget"
    )
    reselect: str = Field(
        default="", validation_alias="hx-reselect", serialization_alias="hx-reselect"
    )
    trigger: str = Field(
        default="", validation_alias="hx-trigger", serialization_alias="hx-trigger"
    )
    trigger_after_settle: str = Field(
        default="",
        validation_alias="hx-trigger-after-settle",
        serialization_alias="hx-trigger-after-settle",
    )
    trigger_after_swap: str = Field(
        default="",
        validation_alias="hx-trigger-after-swap",
        serialization_alias="hx-trigger-after-swap",
    )



from starlette.requests import Request
from typing import Any
from pypox._types import BodyForm


class FormValidator:

    def __init__(self, name: str, request: Request) -> None:
        self._request = request
        self._name = name

    async def __call__(self, _type: type) -> Any:
        if _type not in [BodyForm]:
            return {}
        if self._name.replace("_", "-") in self._request.cookies:
            self._value = await self._request.form()
        else:
            self._value = await self._request.form()
        return {self._name: self.validate(_type)}

    def validate(self, _type: type) -> Any:
        if _type == BodyForm:
            return dict(self._value)
        else:
            raise ValueError("Invalid type")



from starlette.middleware.base import BaseHTTPMiddleware
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTError
from starlette.responses import JSONResponse
import base64


class BearerTokenMiddleware(BaseHTTPMiddleware):

    def __init__(self, app, secret_key: str, algorithm: str, expires_in: int = 3600):
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.expires_in = expires_in
        super().__init__(app)

    async def dispatch(self, request, call_next):
        bearer, token = request.headers.get("Authorization", "").split(" ")
        if bearer.lower() != "bearer":
            return JSONResponse({"detail": "Invalid token type"}, status_code=401)
        try:
            payload = jwt.decode(token, self.secret_key, algorithms=[self.algorithm])
        except ExpiredSignatureError:
            return JSONResponse({"detail": "Token has expired"}, status_code=401)
        except JWTError:
            return JSONResponse({"detail": "Invalid token"}, status_code=401)
        request.state.user = payload
        return await call_next(request)


class BasicTokenMiddleware(BaseHTTPMiddleware):

    def __init__(self, app):
        super().__init__(app)

    async def dispatch(self, request, call_next):
        basic, token = request.headers.get("Authorization", "").split(" ")
        if basic.lower() != "basic":
            return JSONResponse({"detail": "Invalid token type"}, status_code=401)
        username, _, password = base64.b64decode(token).decode().partition(":")
        request.state.user = {"username": username, "password": password}
        return await call_next(request)



from starlette.requests import Request
from typing import Any
from pypox._types import HeaderStr, HeaderInt, HeaderFloat, HeaderBool


class HeaderValidator:

    def __init__(self, name: str, request: Request) -> None:
        self._request = request
        self._name = name

    def __call__(self, _type: type) -> Any:
        if _type not in [HeaderStr, HeaderInt, HeaderFloat, HeaderBool]:
            return {}
        if self._name.replace("_", "-") in self._request.headers:
            self._value = self._request.headers[self._name.replace("_", "-")]
        else:
            self._value = self._request.headers[self._name.replace("-", "_")]
        return {self._name: self.validate(_type)}

    def validate(self, _type: type) -> Any:
        if _type == HeaderStr:
            return str(self._value)
        if _type == HeaderInt:
            return int(self._value)
        if _type == HeaderFloat:
            return float(self._value)
        if _type == HeaderBool:
            return bool(self._value)



from starlette.requests import Request
from typing import Any
from pypox._types import PathStr, PathInt, PathFloat, PathBool


class PathValidator:

    def __init__(self, name: str, request: Request) -> None:
        self._request = request
        self._name = name

    def __call__(self, _type: type) -> Any:
        if _type not in [PathStr, PathInt, PathFloat, PathBool]:
            return
        if self._name.replace("_", "-") in self._request.path_params:
            self._value = self._request.path_params[self._name.replace("_", "-")]
        else:
            self._value = self._request.path_params[self._name.replace("-", "_")]
        return {self._name: self.validate(_type)}

    def validate(self, _type: type) -> Any:
        if _type == PathStr:
            return str(self._value)
        if _type == PathInt:
            return int(self._value)
        if _type == PathFloat:
            return float(self._value)
        if _type == PathBool:
            return bool(self._value)



from functools import wraps
import inspect
from typing import Any, Callable, Awaitable, NewType
from starlette.requests import Request
from starlette.responses import Response
import asyncio
from pypox._types import (
    QueryStr,
    QueryInt,
    QueryFloat,
    QueryBool,
    PathStr,
    PathInt,
    PathFloat,
    PathBool,
    HeaderStr,
    HeaderInt,
    HeaderFloat,
    HeaderBool,
    CookieStr,
    CookieInt,
    CookieFloat,
    CookieBool,
    BodyDict,
    BodyForm,
)
from pypox.processing.form import FormValidator
from pypox.processing.json import JSONValidator
from pypox.processing.query import QueryValidator
from pypox.processing.path import PathValidator
from pypox.processing.header import HeaderValidator
from pypox.processing.cookies import CookieValidator
from pypox.processing.htmx import HTMXHeaders, HTMXValidator


def processor(
    validators: list = [],
) -> Callable:

    def decorator(
        func: Callable,
    ) -> Callable[[Request], Awaitable[Response | Response]]:

        @wraps(func)
        async def wrapper(request: Request) -> Response:
            params = await PypoxProcessor(
                func,
                [
                    QueryValidator,
                    PathValidator,
                    HeaderValidator,
                    CookieValidator,
                    HTMXValidator,
                    JSONValidator,
                    FormValidator,
                ]
                + validators,
            ).validate(
                request,
            )
            if inspect.iscoroutinefunction(func):
                return await func(**params)
            else:
                return func(**params)

        return wrapper

    return decorator


class PypoxProcessor:

    def __init__(
        self,
        func: Callable[[Any], Awaitable[Response] | Response],
        validators: list = [],
    ) -> None:
        self._validators = validators
        self._func = func

    async def validate(self, request: Request) -> Any:
        params = {}
        for name, parameter in self.iterate_params(self._func):
            if parameter.annotation == Request:
                params[name] = request
                break
            for validator in self._validators:
                validator_obj = validator(name, request)
                if inspect.iscoroutinefunction(validator_obj.__call__):
                    data = await validator_obj(parameter.annotation)
                else:
                    data = validator_obj(parameter.annotation)
                if data:
                    params.update(data)
                    await asyncio.sleep(0)
                    break
        return params

    def iterate_params(self, func: Callable[[Any], Awaitable[Response] | Response]):
        for name, parameter in inspect.signature(func).parameters.items():
            yield name, parameter



import os
from typing import (
    Any,
    Awaitable,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    TypeVar,
    ParamSpec,
)
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route, Router, Mount, BaseRoute
from starlette.types import ExceptionHandler, Lifespan, Receive, Scope, Send
from pypox.processing.base import PypoxProcessor, processor
from pypox.router import BaseRouter
from pypox.openapi.main import OpenAPI, Info, License


class Pypox(Starlette):
    """A web application framework based on Starlette.

    Args:
        Starlette (type): The base class for the Pypox application.

    Attributes:
        processor_func (list[BaseProcessor] | None): The list of processor functions.
        conventions (list[BaseConvention]): The list of conventions.
        debug (bool): Flag indicating whether debug mode is enabled.
        middleware (Sequence[Middleware] | None): The middleware stack.
        exception_handlers (Mapping[Any, ExceptionHandler] | None): The exception handlers.
        on_startup (Sequence[Callable[[], Any]] | None): The startup functions.
        on_shutdown (Sequence[Callable[[], Any]] | None): The shutdown functions.
        lifespan (Lifespan | None): The lifespan of the application.
    """

    def __init__(
        self,
        conventions: list[BaseRouter] | None = None,
        debug: bool = False,
        middleware: Sequence[Middleware] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan | None = None,
        open_api_version: str = "3.0.3",
        info: Info = Info(title="Pypox", version="0.0.1", license=License(name="MIT")),
        license: License = License(name="MIT"),
        validators: list[Any] = [],
    ) -> None:
        self._openapi_version = open_api_version
        self._info = info
        self._license = license
        self._validators = validators
        routes: list[BaseRoute] = []
        if conventions:
            for convention in conventions:
                endpoints = convention.generate_routes()
                for endpoint in endpoints:
                    endpoint.endpoint = PypoxProcessor(
                        endpoint.endpoint, validators
                    ).validate
                routes.extend(endpoints)
        super().__init__(
            debug,
            routes,
            middleware,
            exception_handlers,
            on_startup,
            on_shutdown,
            lifespan,
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["openapi"] = {
            "openapi": self._openapi_version,
            "info": self._info,
            "license": self._license,
        }
        return await super().__call__(scope, receive, send)

    def add_route(
        self,
        path: str,
        route: Callable,
        methods: List[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        validators: (
            List[Callable[[Request], Awaitable[Response] | Response]] | None
        ) = None,
    ) -> None:

        self.routes.append(
            Route(
                path,
                processor(validators or [])(route),
                methods=methods,
                name=name,
                include_in_schema=include_in_schema,
            )
        )


class PypoxHTMX(BaseRouter):

    def __init__(
        self,
        directory: str | None = None,
        middleware: Sequence[Middleware] | None = None,
        on_startup: Sequence[Callable[[], Any]] | None = None,
        on_shutdown: Sequence[Callable[[], Any]] | None = None,
        lifespan: Lifespan | None = None,
        open_api_version: str = "3.0.3",
        info: Info = Info(title="Pypox", version="0.0.1", license=License(name="MIT")),
        license: License = License(name="MIT"),
    ) -> None:
        self._openapi_version = open_api_version
        self._info = info
        self._license = license
        super().__init__(
            directory=directory or "", entry_point="page", file={"page.py": "GET"}
        )
        if self._directory:
            self._router = Router(
                self.generate_routes(),
                middleware=middleware,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                lifespan=lifespan,
            )

    async def __call__(
        self,
        scope: MutableMapping[str, Any],
        receive: Callable[[], Awaitable[MutableMapping[str, Any]]],
        send: Callable[[MutableMapping[str, Any]], Awaitable[None]],
    ) -> None:
        if hasattr(self, "_router"):
            return await self._router(scope, receive, send)
        else:
            return await super().__call__(scope, receive, send)
